"""
A credit model that uses a machine learning model to estimate the expected 
value of the the remaining loans in a loan book.

The example shows how to integrate a machine learning model with some simple 
domain information to get contextualized result. It also show cases how statistical
libraries like Seaborn can be used to investigate data before hand. 
"""

from pathlib import Path

import pandas as pd
import seaborn as sns
from sklearn.ensemble import RandomForestClassifier

from fn_graph import Composer

data_path = Path(__file__).parent


def loan_data():
    """
    Load the loan data
    """
    return pd.read_csv(data_path / "credit_data.csv")


def training_data(loan_data):
    """
    Ignore currently live loans
    """
    return loan_data[loan_data.status != "LIVE"]


def investigate_data(training_data):
    """
    Use a seaborn pairplot to get a feel for the data
    """
    return sns.pairplot(training_data.sample(100), hue="status")


def training_features(training_data: pd.DataFrame):
    """
    One hot encode gender and dro =p columns not used in training
    """
    return pd.get_dummies(
        training_data.drop(columns=["outstanding_balance", "status", "account_no"])
    )


def training_target(training_data):
    """
    Convert the target variable to a boolean
    """
    return training_data.status == "DEFAULT"


def model(training_features, training_target):
    """
    Fit a model
    """
    model = RandomForestClassifier()
    model.fit(training_features, training_target)
    return model


def prediction_data(loan_data):
    """
    Only consider the currently live loans
    """
    return loan_data[loan_data.status == "LIVE"]


def prediction_features(prediction_data: pd.DataFrame):
    """
    Prepare the prediction features
    """
    return pd.get_dummies(
        prediction_data.drop(columns=["outstanding_balance", "status", "account_no"])
    )


def probability_of_default(model, prediction_features):
    """
    Predict the probability of default using the trained model
    """
    return model.predict_proba(prediction_features)[:, 1]


def expected_outstanding_repayment(prediction_data, probability_of_default):
    """
    Calculate the expected repayment for each loan
    """
    return prediction_data.assign(probability_of_default=probability_of_default).assign(
        expected_repayment=lambda df: df.outstanding_balance
        * (1 - df.probability_of_default)
    )


def value_of_live_book(expected_outstanding_repayment):
    """
    The total remianing value of the loan book
    """
    return expected_outstanding_repayment.expected_repayment.sum()


composer = Composer().update(
    loan_data,
    training_data,
    training_features,
    training_target,
    investigate_data,
    model,
    prediction_data,
    prediction_features,
    probability_of_default,
    expected_outstanding_repayment,
    value_of_live_book,
)

# Just for uniformity with the rest of the examples
f = composer



"""
A simple machine learning example that builds a classifier for the standard iris dataset.

This example uses scikit-learn to build a a classifier to detect the species of an iris 
flower based on attributes of the flower. Based on the parameters different types of models 
can be trained, and preprocessing can be turned on and off. It also show cases the integration 
of visualisations to measure the accuracy of the model.  
"""

from fn_graph import Composer
import sklearn, sklearn.datasets, sklearn.svm, sklearn.linear_model, sklearn.metrics
from sklearn.model_selection import train_test_split
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pylab as plt


def iris():
    """
    Load the classic iris dataset
    """
    return sklearn.datasets.load_iris()


def data(iris):
    """
    Pull out the data as pandas DataFrame
    """
    df_train = pd.DataFrame(
        iris.data, columns=["feature{}".format(i) for i in range(4)]
    )
    return df_train.assign(y=iris.target)


def investigate_data(data):
    """
    Check for any visual correlations using seaborn
    """
    return sns.pairplot(data, hue="y")


def preprocess_data(data, do_preprocess):
    """
    Preprocess the data by scaling depending on the parameter

    We make sure we don't mutate the data because that is better practice.
    """
    processed = data.copy()
    if do_preprocess:
        processed.iloc[:, :-1] = sklearn.preprocessing.scale(processed.iloc[:, :-1])
    return processed


def split_data(preprocess_data):
    """
    Split the data into test and train sets
    """
    return dict(
        zip(
            ("training_features", "test_features", "training_target", "test_target"),
            train_test_split(preprocess_data.iloc[:, :-1], preprocess_data["y"]),
        )
    )


# This is done verbosely purpose, but it could be more concise


def training_features(split_data):
    return split_data["training_features"]


def training_target(split_data):
    return split_data["training_target"]


def test_features(split_data):
    return split_data["test_features"]


def test_target(split_data):
    return split_data["test_target"]


def model(training_features, training_target, model_type):
    """
    Train the model
    """
    if model_type == "ols":
        model = sklearn.linear_model.LogisticRegression()
    elif model_type == "svm":
        model = sklearn.svm.SVC()
    else:
        raise ValueError("invalid model selection, choose either 'ols' or 'svm'")
    model.fit(training_features, training_target)
    return model


def predictions(model, test_features):
    """
    Make some predictions foo the test data
    """
    return model.predict(test_features)


def classification_metrics(predictions, test_target):
    """
    Show some standard classification metrics
    """
    return sklearn.metrics.classification_report(test_target, predictions)


def plot_confusion_matrix(
    cm, target_names, title="Confusion matrix", cmap=plt.cm.Blues
):
    """
    Plots a confusion matrix using matplotlib. 
    
    This is just a regular function that is not in the composer. 
    Shamelessly taken from https://scikit-learn.org/0.15/auto_examples/model_selection/plot_confusion_matrix.html
    """

    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(target_names))
    plt.xticks(tick_marks, target_names, rotation=45)
    plt.yticks(tick_marks, target_names)
    plt.tight_layout()
    plt.ylabel("True label")
    plt.xlabel("Predicted label")
    return plt.gcf()


def confusion_matrix(predictions, test_target):
    """
    Show the confusion matrix
    """
    cm = sklearn.metrics.confusion_matrix(test_target, predictions)
    return plot_confusion_matrix(cm, ["setosa", "versicolor", "virginica"])


f = (
    Composer()
    .update_parameters(
        # Parameter controlling the model type (ols, svc)
        model_type="ols",
        # Parameter enabling data preprocessing
        do_preprocess=True,
    )
    .update(
        iris,
        data,
        preprocess_data,
        investigate_data,
        split_data,
        training_features,
        training_target,
        test_features,
        test_target,
        model,
        predictions,
        classification_metrics,
        confusion_matrix,
    )
)
